@@ -270,13 +270,30 @@ module Nokogiri
       def add_child node_or_tags
         node_or_tags = coerce(node_or_tags)
         if node_or_tags.is_a?(XML::NodeSet)
-          node_or_tags.each { |n| add_child_node n }
+          node_or_tags.each { |n| add_child_node_and_reparent_attrs n }
         else
-          add_child_node node_or_tags
+          add_child_node_and_reparent_attrs node_or_tags
         end
         node_or_tags
       end
 
+      ###
+      # Add +node_or_tags+ as the first child of this Node.
+      # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
+      #
+      # Returns the reparented node (if +node_or_tags+ is a Node), or NodeSet (if +node_or_tags+ is a DocumentFragment, NodeSet, or string).
+      #
+      # Also see related method +add_child+.
+      def prepend_child node_or_tags
+        if first = children.first
+          # Mimic the error add_child would raise.
+          raise RuntimeError, "Document already has a root node" if document? && !node_or_tags.processing_instruction?
+          first.__send__(:add_sibling, :previous, node_or_tags)
+        else
+          add_child(node_or_tags)
+        end
+      end
+
       ###
       # Add +node_or_tags+ as a child of this Node.
       # +node_or_tags+ can be a Nokogiri::XML::Node, a ::DocumentFragment, a ::NodeSet, or a string containing markup.
@@ -296,7 +313,7 @@ module Nokogiri
       #
       # Also see related method +before+.
       def add_previous_sibling node_or_tags
-        raise ArgumentError.new("A document may not have multiple root nodes.") if parent.is_a?(XML::Document) && !node_or_tags.is_a?(XML::ProcessingInstruction)
+        raise ArgumentError.new("A document may not have multiple root nodes.") if (parent && parent.document?) && !node_or_tags.processing_instruction?
 
         add_sibling :previous, node_or_tags
       end
@@ -309,7 +326,7 @@ module Nokogiri
       #
       # Also see related method +after+.
       def add_next_sibling node_or_tags
-        raise ArgumentError.new("A document may not have multiple root nodes.") if parent.is_a?(XML::Document)
+        raise ArgumentError.new("A document may not have multiple root nodes.") if (parent && parent.document?) && !node_or_tags.processing_instruction?
         
         add_sibling :next, node_or_tags
       end
@@ -361,9 +378,9 @@ module Nokogiri
         node_or_tags = coerce(node_or_tags)
         children.unlink
         if node_or_tags.is_a?(XML::NodeSet)
-          node_or_tags.each { |n| add_child_node n }
+          node_or_tags.each { |n| add_child_node_and_reparent_attrs n }
         else
-          add_child_node node_or_tags
+          add_child_node_and_reparent_attrs node_or_tags
         end
         node_or_tags
       end
@@ -491,6 +508,15 @@ module Nokogiri
       # *this* node.  Returns a XML::NodeSet containing the nodes parsed from
       # +string_or_io+.
       def parse string_or_io, options = nil
+        ##
+        # When the current node is unparented and not an element node, use the
+        # document as the parsing context instead. Otherwise, the in-context
+        # parser cannot find an element or a document node.
+        # Document Fragments are also not usable by the in-context parser.
+        if !element? && !document? && (!parent || parent.fragment?)
+          return document.parse(string_or_io, options)
+        end
+
         options ||= (document.html? ? ParseOptions::DEFAULT_HTML : ParseOptions::DEFAULT_XML)
         if Fixnum === options
           options = Nokogiri::XML::ParseOptions.new(options)
@@ -575,6 +601,16 @@ module Nokogiri
         type == HTML_DOCUMENT_NODE
       end
 
+      # Returns true if this is a Document
+      def document?
+        is_a? XML::Document
+      end
+
+      # Returns true if this is a ProcessingInstruction node
+      def processing_instruction?
+        type == PI_NODE
+      end
+
       # Returns true if this is a Text node
       def text?
         type == TEXT_NODE
@@ -882,7 +918,6 @@ module Nokogiri
         # FIXME: this is a hack around broken libxml versions
         return dump_html if Nokogiri.uses_libxml? && %w[2 6] === LIBXML_VERSION.split('.')[0..1]
 
-        options[:save_with] |= save_option if options[:save_with]
         options[:save_with] = save_option unless options[:save_with]
         serialize(options)
       end
@@ -917,18 +952,23 @@ module Nokogiri
       end
 
       def coerce data # :nodoc:
-        return data                    if data.is_a?(XML::NodeSet)
-        return data.children           if data.is_a?(XML::DocumentFragment)
-        return fragment(data).children if data.is_a?(String)
-
-        if data.is_a?(Document) || data.is_a?(XML::Attr) || !data.is_a?(XML::Node)
-          raise ArgumentError, <<-EOERR
-Requires a Node, NodeSet or String argument, and cannot accept a #{data.class}.
-(You probably want to select a node from the Document with at() or search(), or create a new Node via Node.new().)
-          EOERR
+        case data
+        when XML::NodeSet
+          return data
+        when XML::DocumentFragment
+          return data.children
+        when String
+          return fragment(data).children
+        when Document, XML::Attr
+          # unacceptable
+        when XML::Node
+          return data
         end
 
-        data
+        raise ArgumentError, <<-EOERR
+Requires a Node, NodeSet or String argument, and cannot accept a #{data.class}.
+(You probably want to select a node from the Document with at() or search(), or create a new Node via Node.new().)
+        EOERR
       end
 
       def implied_xpath_context
@@ -938,6 +978,15 @@ Requires a Node, NodeSet or String argument, and cannot accept a #{data.class}.
       def inspect_attributes
         [:name, :namespace, :attribute_nodes, :children]
       end
+
+      def add_child_node_and_reparent_attrs node
+        add_child_node node
+        node.attribute_nodes.find_all { |a| a.name =~ /:/ }.each do |attr_node|
+          attr_node.remove
+          node[attr_node.name] = attr_node.value
+        end
+      end
+
     end
   end
 end
